package repub;

import java.util.List;
import java.util.Random;
import java.util.Vector;

/**
 * Class for Anonymized Group
 * 
 * @author Yingyi Bu
 * 
 */
public class AGroup {

	/**
	 * list of individuals in the anonymized group
	 */
	private List individuals = new Vector();

	private int start = 0;

	private int end = 0;

	public AGroup() {

	}

	/**
	 * anonymized group
	 * 
	 * @param gn
	 * @param ins
	 * @param s
	 * @param e
	 */
	public AGroup(List ins, int s, int e) {
		individuals = ins;
		start = s;
		end = e;
	}

	/**
	 * new an anonymized group
	 * 
	 * @param ins
	 */
	public AGroup(List ins) {
		individuals = ins;
	}

	/**
	 * update the signatures of an individual
	 */
	public void updateSignatures(Generalization gn) {
		Random rd = new Random(System.currentTimeMillis());

		/**
		 * set up all holders
		 */
		for (int i = 0; i < Constant.num_of_ps; i++) {
			for (int j = 0; j < individuals.size(); j++) {
				Individual person = (Individual) individuals.get(j);
				if (person.isContract(new Integer(i))) {
					person.setRole(gn, i, Constant.HOLDER);
					gn.putIntoCohort(person, i, 0);
				}
			}
		}

		int position = 1;
		/**
		 * process for each permanent sensitive values
		 */
		for (int i = 0; i < Constant.num_of_ps; i++) {
			/**
			 * test the number of holders and decoys
			 */
			int holders = this.countHolder(i);
			int decoys = this.countDecoy(i);
			while (holders * (Constant.L - 1) > decoys) {
				/**
				 * randomly select decoys from an AG firstly get a list of
				 * candidates
				 */
				List candidates = this.getCleanOrBuddy(i);
				int size = candidates.size();

				if (size == 0) {
					System.out.println("no clean one for decoys! ");
					break;
				}
				/**
				 * randomly select one as decoys
				 */
				int selection = Math.abs(rd.nextInt()) % size;
				/**
				 * get this individual
				 */
				Individual person = (Individual) candidates.get(selection);
				System.out.println("selected decoy for " + i + ": " + person
						+ " put in cohort " + position);
				/**
				 * set this person's role as a decoy
				 */
				person.setRole(gn, i, Constant.DECOY, person.getRole(i));

				gn.putIntoCohort(person, i, position);
				position = (position + 1) % Constant.L;
				if (position == 0)
					position = position + 1;
				/**
				 * count decoy and holder again
				 */
				// holders = this.countHolder(i);
				decoys++;
			}
		}
	}

	/**
	 * get all the dis-clean individuals
	 * 
	 * @param dis
	 * @return
	 */
	public List getCleanOrBuddy(int dis) {
		List clean = new Vector();
		for (int i = 0; i < individuals.size(); i++) {
			Individual person = (Individual) this.individuals.get(i);
			if (person.getRole(dis) == Constant.CLEAN
					|| person.getRole(dis) == Constant.BUDDY)
				clean.add(person);
		}
		return clean;
	}

	/**
	 * count the number of holders
	 * 
	 * @param dis
	 * @return
	 */
	public int countHolder(int dis) {

		int holder = 0;
		for (int i = 0; i < individuals.size(); i++) {
			Individual person = (Individual) this.individuals.get(i);
			int role = person.getRole(dis);
			if (role == Constant.HOLDER)
				holder++;
		}
		return holder;
	}

	/**
	 * count the number of decoys
	 * 
	 * @param dis
	 *            the disease
	 * @return
	 */
	public int countDecoy(int dis) {
		int decoy = 0;
		for (int i = 0; i < individuals.size(); i++) {
			Individual person = (Individual) this.individuals.get(i);
			int role = person.getRole(dis);
			if (role == Constant.DECOY)
				decoy++;
		}
		return decoy;
	}

	/**
	 * return a list of tuples which are generalized version
	 * 
	 * @return
	 */
	public List toTuples() {
		float lower[] = getLower();
		float upper[] = getUpper();

		List tuples = new Vector();

		/**
		 * iterate all the individuals to get the generalized tuples
		 */
		for (int i = 0; i < individuals.size(); i++) {
			Individual person = (Individual) individuals.get(i);
			List disease = person.getDisease();
			for (int j = 0; j < disease.size(); j++) {
				StringBuffer tuple = new StringBuffer("");
				tuple.append(start + "-" + end + " ");
				tuple.append(person.getPID() + " ");
				/**
				 * set all the interval of the tuple
				 */
				for (int k = 0; k < Constant.DIM; k++) {
					tuple.append(lower[k] + "-" + upper[k] + " ");
				}
				/**
				 * get the sensitive value
				 */
				String dis = (String) SensitiveValue.diseaseIDToName
						.get(disease.get(j));

				tuple.append(dis);

				/*
				 * for (int s = 0; s < Constant.num_of_ps; s++) { if
				 * (person.getRole(s) == Constant.DECOY) { tuple.append(" decoy " +
				 * s + " "); } }
				 */
				tuples.add(tuple);
			}
		}
		return tuples;
	}

	/**
	 * get the lower bound values of QIDs of Anonymized Group
	 * 
	 * @return the lower bound of the array
	 */
	public float[] getLower() {
		float lower[][] = new float[individuals.size()][];
		for (int i = 0; i < individuals.size(); i++) {
			Individual in = (Individual) individuals.get(i);
			lower[i] = in.getQID();
		}

		float value[] = new float[Constant.DIM];
		for (int i = 0; i < value.length; i++)
			value[i] = 9999999;
		for (int j = 0; j < lower.length; j++) {
			for (int k = 0; k < Constant.DIM; k++) {
				if (value[k] > lower[j][k])
					value[k] = lower[j][k];
			}
		}
		return value;
	}

	/**
	 * get the upper values of QIDs of an Anonymized Group
	 * 
	 * @return
	 */
	public float[] getUpper() {
		float upper[][] = new float[individuals.size()][];
		for (int i = 0; i < individuals.size(); i++) {
			Individual in = (Individual) individuals.get(i);
			upper[i] = in.getQID();
		}

		float value[] = new float[Constant.DIM];
		for (int i = 0; i < value.length; i++)
			value[i] = 0;
		for (int j = 0; j < upper.length; j++) {
			for (int k = 0; k < Constant.DIM; k++) {
				if (value[k] < upper[j][k])
					value[k] = upper[j][k];
			}
		}
		/**
		 * return the upper bound
		 */
		return value;

	}

	public static void main(String args[]) {
		float[] qid1 = { 2, 3, 5 };
		float[] qid2 = { 1, -2, 6 };
		float[] qid3 = { 8, 100, 4 };
		Individual in1 = new Individual(1, qid1);
		Individual in2 = new Individual(2, qid2);
		Individual in3 = new Individual(3, qid3);

		List ins = new Vector();

		ins.add(in1);
		ins.add(in2);
		ins.add(in3);

		AGroup ag = new AGroup(ins);

		float lower[] = ag.getLower();

		float upper[] = ag.getUpper();

		System.out.println("lower bound: ");
		for (int i = 0; i < lower.length; i++)
			System.out.println(lower[i] + "  ");
		System.out.println();

		System.out.println("upper bound");
		for (int i = 0; i < upper.length; i++)
			System.out.println(upper[i] + "  ");
		System.out.println();
	}

	public int getStart() {
		return start;
	}

	public int getEnd() {
		return end;
	}

	/**
	 * merge two anonymized groups return the fresh group they get
	 * 
	 * @param ag
	 */
	public static AGroup merge(AGroup group, AGroup ag) {
		AGroup freshGroup = new AGroup();
		List list = new Vector();
		list.addAll(group.getIndividuals());
		list.addAll(ag.getIndividuals());
		/**
		 * set a fresh anonymized group
		 */
		freshGroup.setIndividuals(list);
		/**
		 * return the merged new group
		 */
		return freshGroup;
	}

	public void setIndividuals(List ls) {
		individuals = ls;
	}

	public List getIndividuals() {
		return individuals;
	}

	/**
	 * get the higher values of Anonymized Group
	 * 
	 * @return
	 */
	public float[] getContainerLower() {
		float lower[][] = new float[individuals.size()][];
		for (int i = 0; i < individuals.size(); i++) {
			Individual in = (Individual) individuals.get(i);
			lower[i] = in.getContainerQIDLower();
		}

		float value[] = new float[Constant.DIM];
		for (int i = 0; i < value.length; i++)
			value[i] = 9999999;
		for (int j = 0; j < lower.length; j++) {
			for (int k = 0; k < Constant.DIM; k++) {
				if (value[k] > lower[j][k])
					value[k] = lower[j][k];
			}
		}
		return value;
	}

	/**
	 * get the lower values of Anonymized Group
	 * 
	 * @return
	 */
	public float[] getContainerUpper() {
		float upper[][] = new float[individuals.size()][];
		for (int i = 0; i < individuals.size(); i++) {
			Individual in = (Individual) individuals.get(i);
			upper[i] = in.getContainerQIDUpper();
		}

		float value[] = new float[Constant.DIM];
		for (int i = 0; i < value.length; i++)
			value[i] = 0;
		for (int j = 0; j < upper.length; j++) {
			for (int k = 0; k < Constant.DIM; k++) {
				if (value[k] < upper[j][k])
					value[k] = upper[j][k];
			}
		}
		/**
		 * return the upper bound
		 */
		return value;
	}

}
