package repub;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Random;
import java.util.Vector;

public class Generalization {

	PrintWriter log = null;

	/**
	 * map between people ID and the individual
	 */
	HashMap voters = new HashMap();

	/**
	 * registration list
	 */
	List rlist = new Vector();

	List genList = new Vector();

	/**
	 * a list of present individuals
	 */
	List present = new Vector();

	/**
	 * a list of absent individuals
	 */
	List absent = new Vector();

	/**
	 * the list of people involved
	 */
	List peopleList = new Vector();

	/**
	 * the pool to store currently removed individuals
	 */
	List loserPool = new Vector();

	/**
	 * cohort information
	 */
	ICohort[][] cohorts = null;

	/**
	 * store all the anonymized groups
	 */
	private List groups = new Vector();

	/**
	 * generate the files as output anonymized tables
	 */
	PrintWriter fileGenerator = null;

	/**
	 * the mapping between disease and count
	 */
	HashMap diseaseToCount = new HashMap();

	/**
	 * show if the percentage should be printed
	 */
	boolean printPercentage = false;

	/**
	 * number of persons whose sensitive values are perturbed
	 */
	public static int replace = 0;

	/**
	 * the construction function
	 * 
	 * @param target,
	 *            the target file of output
	 * 
	 * 
	 * Initialize the cohort information
	 */
	public Generalization(String target, int timestamp) {

		/**
		 * if cohorts = initialize num_of_ps Legion
		 */
		cohorts = new ICohort[Constant.num_of_ps][];
		/**
		 * initialize each cohort array
		 */
		for (int i = 0; i < Constant.num_of_ps; i++) {
			/**
			 * new a cohort array
			 */
			cohorts[i] = new ICohort[Constant.L];
			for (int j = 0; j < Constant.L; j++) {
				/**
				 * new a cohort N.O. j for permanent sensitive value i
				 */
				cohorts[i][j] = new ICohort(i, j);
			}
			try {
				log = new PrintWriter(new OutputStreamWriter(
						new FileOutputStream("log/log" + timestamp + ".txt")));
				/**
				 * initialize a file generator to generate files
				 */
				fileGenerator = new PrintWriter(new OutputStreamWriter(
						new FileOutputStream(target)));
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * print an anonymized group to the files
	 * 
	 * @param ag
	 *            AGroup, to be printed
	 */
	public void printGroup(AGroup ag) {

		List tuples = ag.toTuples();
		/**
		 * print all the tuples
		 */
		for (int j = 0; j < tuples.size(); j++)
			fileGenerator.println(tuples.get(j));
	}

	/**
	 * put a person into a cohort: cohort_no
	 * 
	 * @param person
	 * @param no
	 */
	public void putIntoCohort(Individual person, int dis, int no) {
		/**
		 * if the person is not in the cohort, put it in
		 */
		if (cohorts[dis][no] != null && !cohorts[dis][no].contain(person)) {
			cohorts[dis][no].put(person);
			person.setCohort(cohorts[dis][no], dis);
		} else {
			System.out.println("null cohort!");
		}
	}

	/**
	 * update the signature
	 * 
	 * @param ag
	 */
	public void updateSignatures(AGroup ag) {
		ag.updateSignatures(this);
	}

	/**
	 * top down partition for the list in the function, a recursive partition
	 * function is called
	 */
	public void topDownPartition() {

		/**
		 * sort the individual by a hilbert curve on their QID attributes
		 */
		hilbertPartition();

		/**
		 * divide the data into part portions, and anonymize each portion
		 */
		int division = peopleList.size() / Constant.part;

		/**
		 * start the top-down specialization
		 */
		System.out.println("start top down specialization: ");
		System.out.println("division: " + division);
		/**
		 * top down specialization each portion
		 */
		for (int i = 0; i <= Constant.part; i++) {
			if (i * division + division - 1 <= peopleList.size() - 1) {
				topDownPartition(peopleList, i * division, i * division
						+ division - 1);
			} else
				topDownPartition(peopleList, i * division,
						peopleList.size() - 1);
		}

		System.out.println("Start clique checking");
		/**
		 * after partition check if constrains of cliques are satisfied
		 */
		checkClique();

		/**
		 * print each group and update their signatures
		 */
		for (int i = 0; i < groups.size(); i++) {
			AGroup ag = (AGroup) groups.get(i);
			/**
			 * update the signatures of the group
			 */
			this.updateSignatures(ag);
			this.printGroup(ag);
		}

		/**
		 * print all the cohorts
		 */
		printCohorts();

		fileGenerator.flush();
		System.out.println("Finish Generalization!");
		/**
		 * writer the udpated individual list and cohorts into disk
		 */
		try {
			ObjectOutputStream writer = new ObjectOutputStream(
					new FileOutputStream(Constant.internalList));
			writer.writeObject(rlist);
			writer.writeObject(voters);
			/**
			 * cohorts
			 */
			writer.writeObject(cohorts);

			/**
			 * people list is written to the disk
			 */
			writer.writeObject(peopleList);
			writer.flush();
			fileGenerator.flush();
			fileGenerator.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Partition the data by hilbert curve
	 * 
	 */
	private void hilbertPartition() {
		/**
		 * clear the list of individuals, and regenerate the hilbert code for
		 * each individual
		 */
		for (int i = 0; i < peopleList.size(); i++) {
			Individual p = (Individual) peopleList.get(i);

			int code = Hilbert.doHilbert(p);
			/**
			 * set each individuals' hilbert code
			 */
			p.setHashValue(code);
		}

		System.out.println("start sort");

		/**
		 * sort the rlist by hilbert curve
		 */
		Collections.sort(peopleList);

		for (int i = 0; i < peopleList.size(); i++) {
			Individual p = (Individual) peopleList.get(i);

			int code = Hilbert.doHilbert(p);
			/**
			 * set each individuals' hilbert code
			 */
			p.setHashValue(code);
		}

		System.out.println("end sort");
	}

	/**
	 * check if clique is violated
	 * 
	 */
	public void checkClique() {

		for (int i = 0; i < peopleList.size(); i++) {
			Individual p = (Individual) peopleList.get(i);
			// p.clearDisease();
			/**
			 * set the person to be not available
			 */
			p.setAvailble(true);
		}

		/**
		 * for each permanent sensitive value, form a clique
		 */
		for (int i = 0; i < Constant.num_of_ps; i++) {
			/**
			 * form the clique for disease i
			 */
			Clique[] cliques = formCliques(i);

			if (cliques != null) {
				/**
				 * merge cliques posed by constaints of clique
				 */
				for (int j = 0; j < cliques.length; j++) {
					/**
					 * for each clique, prepare to merge transform it into an
					 * anonymized group
					 */
					AGroup ag = new AGroup(cliques[j].getIndividuals());
					/**
					 * get the number of groups
					 */
					int size = groups.size();
					/**
					 * set the new group
					 */
					for (int k = 0; k < size; k++) {
						/**
						 * enumerate all current anonymized groups
						 */
						AGroup group = (AGroup) groups.get(k);
						/**
						 * get the group people list
						 */
						List groupPeople = group.getIndividuals();
						/**
						 * get the clique people list
						 */
						List cliquePeople = cliques[j].getIndividuals();

						boolean overlap = false;
						for (int l = 0; l < groupPeople.size(); l++) {
							if (cliquePeople.contains(groupPeople.get(l))) {
								overlap = true;
								break;
							}
						}
						if (overlap == true) {
							ag = AGroup.merge(ag, group);
							// System.out.println("merge groups");
							groups.remove(group);
						}
					}
					groups.add(ag);

					/**
					 * print the formed group that contains a clique
					 */
					List tuples = ag.toTuples();
					/**
					 * print all the tuples
					 */
					for (int n = 0; n < tuples.size(); n++)
						System.out.println(tuples.get(j));
				}
			}
		}

		System.out.println("check cohort size");
		// /**
		// * put the person into the cohort
		// */
		// for (int i = 0; i < peopleList.size(); i++) {
		// Individual person = (Individual) peopleList.get(i);
		// for (int j = 0; j < Constant.num_of_ps; j++) {
		// if (person.getRole(j) == Constant.HOLDER) {
		// if (!cohorts[j][0].contain(person))
		// cohorts[j][0].put(person);
		// }
		// }
		// }

		/**
		 * check whether cohorts are of the same size
		 */
		for (int i = 0; i < Constant.num_of_ps; i++) {
			boolean equal = true;
			int size = cohorts[i][0].size;
			for (int j = 0; j < Constant.L; j++) {
				if (cohorts[i][j].size != size)
					equal = false;
				if (j == 0) {
					List patients = cohorts[i][j].getIndividuals();
					for (int k = 0; k < patients.size(); k++) {
						Individual per = (Individual) patients.get(k);
						if (per.getRole(i) != Constant.HOLDER) {
							equal = false;
							System.out.println("XXXXX holder cohort error!");
						}
					}
				} else {
					List patients = cohorts[i][j].getIndividuals();
					for (int k = 0; k < patients.size(); k++) {
						Individual per = (Individual) patients.get(k);
						if (per.getRole(i) != Constant.DECOY) {
							equal = false;
							System.out.println("XXXXX decoy cohort error!");
						}
					}
				}
			}
			if (equal == false)
				System.out.println("Cohort error!");
			else
				System.out.println("Cohorts are good");
		}
	}

	/**
	 * print all the cohort information
	 * 
	 */
	private void printCohorts() {
		/**
		 * print all the cohorts
		 */
		for (int i = 0; i < Constant.num_of_ps; i++) {
			System.out.println("disease " + i + ": ");
			for (int j = 0; j < Constant.L; j++) {
				// System.out.println("cohort " + j + ": ");
				List ins = cohorts[i][j].getIndividuals();
				for (int k = 0; k < ins.size(); k++)
					System.out.println(ins.get(k));
			}
		}
	}

	/**
	 * top down partition the individual list
	 * 
	 * @param individuals
	 * @param start
	 * @param end
	 */
	public void topDownPartition(List individuals, int start, int end) {

		if (start > end)
			return;

		boolean canPartition = false;
		int division = (end - start + 1) / Constant.BRUNCH;
		if (division >= Constant.L) {
			for (int i = 0; i < Constant.BRUNCH - 1; i++) {
				canPartition = satisfy(individuals, start + i * division, start
						+ i * division + division - 1);
				if (canPartition == false)
					break;
			}
			if (canPartition == true)
				canPartition = satisfy(individuals, start
						+ (Constant.BRUNCH - 1) * division, end);
		}
		if (canPartition) {
			for (int i = 0; i < Constant.BRUNCH - 1; i++) {
				topDownPartition(individuals, start + i * division, start + i
						* division + division - 1);
			}
			topDownPartition(individuals, start + (Constant.BRUNCH - 1)
					* division, end);
		} else {
			List ins = new Vector();
			for (int i = start; i <= end; i++)
				ins.add(individuals.get(i));
			AGroup ag = new AGroup(ins, start, end);
			groups.add(ag);
		}
	}

	/**
	 * check whether the individuals from start to end satisfy the privacy
	 * requirement
	 * 
	 * @param individuals,
	 *            a list of individuals
	 * @param start,
	 *            the start position
	 * @param end,
	 *            the end position
	 * @return the boolean variable true or false
	 */
	public boolean satisfy(List individuals, int start, int end) {

		if (end - start < Constant.L)
			return false;
		System.gc();

		/**
		 * disease to count
		 */
		diseaseToCount.clear();

		List dis = new Vector();

		for (int i = start; i <= end; i++) {
			Individual person = (Individual) individuals.get(i);
			List disease = person.getDisease();
			for (int j = 0; j < disease.size(); j++) {
				Integer diseaseID = (Integer) disease.get(j);
				/**
				 * set the count of the disease
				 */
				if (diseaseToCount.get(diseaseID) == null) {
					/**
					 * add a new disease with count 1
					 */
					diseaseToCount.put(diseaseID, new Count());
					dis.add(diseaseID);
				} else {
					Count count = (Count) diseaseToCount.get(diseaseID);
					count.increaseOne();
				}
			}
		}

		int total = end - start + 1;

		float criteria = (float) 1 / (float) Constant.L;
		for (int i = 0; i < dis.size(); i++) {
			Integer diseaseID = (Integer) dis.get(i);
			Count count = (Count) diseaseToCount.get(diseaseID);
			float percentage = (float) count.count / (float) total;
			if (percentage > criteria) {
				return false;
			}
		}
		return true;
	}

	/**
	 * form all the cliques of a disease
	 * 
	 * @param dis
	 * @return
	 */
	public Clique[] formCliques(int dis) {

		int size = cohorts[dis][0].size;
		if (size == 0)
			return null;

		cliqueExtinct(dis);

		/*
		 * get the cohorts
		 */

		/**
		 * the number of holder = the number of clique
		 */
		Clique[] cliques = new Clique[size];

		/**
		 * size is the total number
		 */
		int round = 0;

		/**
		 * find the clique
		 */
		for (; round < size; round++) {
			/**
			 * randomly find the starting clique
			 */
			Random rd = new Random(System.currentTimeMillis());
			int start = rd.nextInt() % Constant.L;
			/**
			 * get the clique
			 */
			Clique newClique = getCliqueFrom(dis, start);

			/**
			 * add the clique into results;
			 */
			cliques[round] = newClique;
		}
		return cliques;
	}

	/**
	 * clique extinction
	 * 
	 * @param dis
	 */
	private void cliqueExtinct(int dis) {

		/**
		 * get the size of each cohort
		 */
		int size = cohorts[dis][0].size;
		/**
		 * an array of list
		 */
		List[] removable = new List[Constant.L];

		/**
		 * initialze a list of removable individuals
		 */
		for (int i = 0; i < Constant.L; i++) {
			removable[i] = new Vector();
		}

		int minsize = 999999;

		/**
		 * select extinct indiivduals
		 */
		for (int i = 0; i < Constant.L; i++) {
			List candidates = cohorts[dis][i].getIndividuals();
			for (int j = 0; j < size; j++) {
				Individual person = (Individual) candidates.get(j);
				if (person.isExtinct())
					removable[i].add(person);
			}
			/**
			 * set up the minmum size of cohorts
			 */
			if (removable[i].size() < minsize) {
				minsize = removable[i].size();
			}
		}

		/**
		 * remove all the extincted cliques
		 */
		for (int i = 0; i < Constant.L; i++) {
			for (int j = 0; j < minsize; j++) {
				Individual person = (Individual) removable[i].get(j);
				cohorts[dis][i].removeIndividual(person);
			}
		}

		System.out.println(minsize + " cliques extinct");
	}

	/**
	 * get a clique from the start cohort
	 * 
	 * @param start
	 * @return
	 */
	public Clique getCliqueFrom(int dis, int start) {

		List[] buckets = new List[Constant.L];

		/**
		 * get the L individual List of each bucket of disease dis
		 */
		for (int j = 0; j < buckets.length; j++) {
			buckets[j] = cohorts[dis][j].getIndividuals();
		}

		int size = buckets[start].size();

		Individual searchStart = null;
		/**
		 * find the first availabe one in List(start)
		 */
		for (int i = 0; i < size; i++) {
			/**
			 * get the individual
			 */
			Individual person = (Individual) buckets[start].get(i);
			if (person.isAvailable() == true) {
				searchStart = person;
				break;
			}
		}

		if (searchStart == null) {
			// System.out.println("clique forming error!");
			return null;
		}

		Clique clique = new Clique();
		log.println("start individual: " + searchStart);

		/**
		 * add search start into the clique
		 */
		clique.addIndividual(searchStart);
		/**
		 * start search for clique
		 */
		for (int i = 0; i < buckets.length; i++) {
			if (i == start)
				continue;
			clique.addIndividual(findNN(searchStart, buckets[i]));
		}
		// System.out.println("Clique formed of Size " + clique.size());

		try {
			log.println("Clique formed of Size " + clique.size());
			log.flush();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return clique;
	}

	/**
	 * find the nn of searchstart in a bucket
	 * 
	 * @param searchStart
	 * @param people
	 * @return
	 */
	public Individual findNNContainer(Individual searchStart, List people) {
		int size = people.size();
		float minDist = Integer.MAX_VALUE;
		int position = -1;
		for (int i = 0; i < size; i++) {
			Individual candidate = (Individual) people.get(i);
			/**
			 * get the dist between candidate and the individual
			 */
			float dist = searchStart.distance(candidate);
			if (dist < minDist) {
				minDist = dist;
				position = i;
			}
		}

		if (position == -1)
			return null;
		Individual result = (Individual) people.get(position);

		// result.setAvailble(false);
		return result;
	}

	/**
	 * find the nn of searchstart in a bucket
	 * 
	 * @param searchStart
	 * @param people
	 * @return
	 */
	public Individual findNN(Individual searchStart, List people) {
		int size = people.size();
		float minDist = Integer.MAX_VALUE;
		int position = -1;
		for (int i = 0; i < size; i++) {
			Individual candidate = (Individual) people.get(i);
			if (candidate.isAvailable() == false)
				continue;
			/**
			 * get the dist between candidate and the individual
			 */
			float dist = searchStart.distance(candidate);
			if (dist < minDist) {
				minDist = dist;
				position = i;
			}
		}
		Individual result = (Individual) people.get(position);

		// result.setAvailble(false);
		return result;
	}

	/**
	 * return the distortion value: the volumne
	 * 
	 * @return
	 */
	public double distortion() {
		double area = 0;
		for (int i = 0; i < groups.size(); i++) {
			AGroup ag = (AGroup) groups.get(i);
			float[] lower = ag.getContainerLower();
			float[] upper = ag.getContainerUpper();
			double region = 1;
			boolean abnormal = false;
			for (int j = 0; j < lower.length; j++) {
				if (upper[j] - lower[j] >= 80) {
					abnormal = true;
					break;
				}
				region = region * (upper[j] - lower[j]);
			}
			if (abnormal == true)
				abnormal = false;
			else
				area += region;
		}

		return area;
	}

	/**
	 * Load data and print the cohorts
	 * 
	 */
	public void loadRLData() {

		try {
			ObjectInputStream input = new ObjectInputStream(
					new FileInputStream(Constant.internalList));
			rlist = (List) input.readObject();

			voters = (HashMap) input.readObject();

			cohorts = (ICohort[][]) input.readObject();

			peopleList = (List) input.readObject();

			this.hilbertPartition();

			refreshRL();
			/**
			 * let
			 */
			input.close();

		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * refesh the RL data
	 * 
	 */
	private void refreshRL() {
		for (int i = 0; i < Constant.num_of_ps; i++) {
			log.println("disease " + i);
			/**
			 * for L bucket, print the individuals
			 */
			for (int j = 0; j < Constant.L; j++) {
				List individuals = cohorts[i][j].getIndividuals();
				for (int k = 0; k < individuals.size(); k++)
					log.println(individuals.get(k));
			}
		}
		log.flush();

		AGroup ag = new AGroup(rlist);
		float[] lower = ag.getLower();
		float[] upper = ag.getUpper();

		Random rd = new Random(System.currentTimeMillis());

		/**
		 * generate new RL QID change rate is 1%
		 */
		for (int i = 0; i < rlist.size(); i++) {
			Individual person = (Individual) rlist.get(i);
			int change = Math.abs(rd.nextInt()) % 100;
			if (change == 0) {
				float[] QID = new float[Constant.DIM];
				for (int j = 0; j < QID.length; j++) {
					QID[j] = lower[j] + rd.nextFloat() * (upper[j] - lower[j])
							/ 20;
				}
				person.setQID(QID);

				/**
				 * replace person's role if any
				 */
				replaceRole(person);
			}

			/**
			 * randomly decide if print the tuple only 0.2% people are removed,
			 * not included in peopleList
			 */
			int put = Math.abs(rd.nextInt()) % 500;
			if (put == 0) {
				/**
				 * replace person's role if any
				 */
				replaceRole(person);
				/**
				 * not included in people list
				 */
				peopleList.remove(person);
				loserPool.add(person);
			}

			/**
			 * determine the restored individuals
			 */
			if (put == 1 && loserPool.size() > 0) {
				if (loserPool.size() >= 1) {
					/**
					 * restore the loser
					 */
					Individual p = (Individual) loserPool.get(0);
					/**
					 * put p into the list
					 */
					if (p != null) {
						peopleList.add(person);
						loserPool.remove(p);
					}
				}
			}

			/**
			 * determine the extinct perons
			 */
			if (put == 2) {
				person.setExtinct(true);

				boolean dirty = false;
				for (int s = 0; s < Constant.num_of_ps; s++)
					if (person.getRole(s) == Constant.HOLDER
							|| person.getRole(s) == Constant.DECOY) {
						dirty = true;
					}
				/**
				 * if person is clean and extinct, then remove it
				 */
				if (!dirty)
					peopleList.remove(person);
			}
		}
	}

	/**
	 * replace the role of the decoy or holder when their role change
	 * 
	 * @param person
	 */
	public void replaceRole(Individual person) {
		for (int i = 0; i < Constant.num_of_ps; i++) {
			if (person.getRole(i) == Constant.HOLDER
					|| person.getRole(i) == Constant.DECOY) {
				SContainer cn = person.getContainer(i);
				/**
				 * remove person from the container
				 */
				person.setContainer(null, i);
				cn.remove(person);
				/**
				 * get the substitution individual
				 */
				Individual sub = cn.getNextBuddy(i);
				/**
				 * if container is not empty, and a decoy or holder leave
				 * container a buddy will take place of the decoy or holder's
				 * role
				 */
				if (sub != null) {
					/**
					 * set substituion's role as the person
					 */
					sub.setRole(this, i, person.getRole(i), sub.getRole(i));
					/**
					 * if person is a holder, pin the sensitive value to the
					 * substitution
					 */
					if (person.getRole(i) == Constant.HOLDER) {
						/**
						 * perturb the disease i to the substitution person
						 */
						sub.addDisease(new Integer(i));
						person.removeDisease(new Integer(i));
						/**
						 * the count of replace increased by one
						 */
						replace++;
					}
					/**
					 * set person as clean
					 */
					person.setRole(this, i, Constant.CLEAN, person.getRole(i));

					/**
					 * change the cohort; put sub in, remove person out
					 */
					ICohort cohort = person.getCohort(i);
					cohort.removeIndividual(person);
					person.setCohort(null, i);

					/**
					 * sub into the cohort
					 */
					cohort.put(sub);
					sub.setCohort(cohort, i);

				} else {
					/**
					 * the container will be empty, if the holder or decoy
					 * removed
					 */
					if (person.getRole(i) == Constant.DECOY
							&& person.isContract(new Integer(i))) {
						/**
						 * randomly select an i-clean individual
						 */
						Individual perturb = randomSelectClean(peopleList, i);
						/**
						 * perturb the sensitive value i
						 */
						person.removeDisease(new Integer(i));
						perturb.addDisease(new Integer(i));

						/**
						 * set the role of the pertubed individual
						 */
						perturb.setRole(this, i, person.getRole(i),
								Constant.CLEAN);
						/**
						 * set person as clean
						 */
						person.setRole(this, i, Constant.CLEAN, person
								.getRole(i));

						ICohort cohort = person.getCohort(i);
						/**
						 * remove person out of the cohort
						 */
						cohort.removeIndividual(person);
						person.setCohort(null, i);
						/**
						 * put the new individual into the cohort
						 */
						cohort.put(perturb);
						perturb.setCohort(cohort, i);

						/**
						 * count of perturbtion increased by one
						 */
						replace++;

					} else {
						log.println("buddy aboundance assumption violated!");
					}
					person.setContainer(cn, i);
					cn.addInidividual(person);
				}
			}
		}

	}

	/**
	 * randomly select some dis-clean individuals from people list
	 * 
	 * @param people
	 * @param dis
	 * @return
	 */
	public Individual randomSelectClean(List people, int dis) {
		Random rand = new Random(System.currentTimeMillis());
		Individual person = null;
		do {
			int pos = rand.nextInt() % people.size();
			person = (Individual) people.get(pos);
		} while (person.getRole(dis) != Constant.CLEAN);
		/**
		 * return the s-clean individual
		 */
		return person;
	}

	/**
	 * read the initial RL information
	 * 
	 * @param dataset
	 */
	public void readStartRLData(String dataset) {
		try {
			String tuple = "";
			int size = 0;
			BufferedReader reader = new BufferedReader(new InputStreamReader(
					new FileInputStream(dataset)));
			while ((tuple = reader.readLine()) != null)
				size++;

			// System.out.println(size);
			reader.close();
			System.gc();

			// String[] results = new String[size];
			reader = new BufferedReader(new InputStreamReader(
					new FileInputStream(dataset)));
			int i = 0;
			/**
			 * read in all the tuples
			 */
			int count = 0;
			while ((tuple = reader.readLine()) != null && i < size) {
				// results[i] = tuple;
				Individual p = parseRLData(tuple);

				if (p != null) {
					// System.out.println(p);
					count++;
				}
				i++;
			}
			System.out.println("total people" + count);
			reader.close();

			ObjectOutputStream writer = new ObjectOutputStream(
					new FileOutputStream(Constant.internalList));
			writer.writeObject(rlist);
			writer.writeObject(voters);
			writer.writeObject(cohorts);
			writer.writeObject(peopleList);
			writer.flush();
			// return results;
		} catch (Exception e) {
			e.printStackTrace();
			// return null;
		}

	}

	/**
	 * parse RL file into a list of Individuals
	 * 
	 * @param person
	 * @return
	 */
	public Individual parseRLData(String person) {
		Random rd = new Random(System.currentTimeMillis());
		/**
		 * setup the voter's list
		 */
		// for (int i = 0; i < people.length; i++) {
		String[] fields = person.split("!");
		// for (int i = 0; i < fields.length; i++)
		// System.out.print(fields[i] + " ");
		// System.out.println();

		try {
			if (fields[0].equals(""))
				return null;
			int pid = Integer.parseInt(fields[0]);
			if (fields[4].equals("-") || fields[6].equals("-")
					|| fields[7].equals("-"))
				return null;
			float age = Float.parseFloat(fields[3]);
			float weight = Float.parseFloat(fields[6]);
			float height = Float.parseFloat(fields[7]);
			float[] qid = new float[Constant.DIM];

			qid[0] = age;
			qid[1] = weight;
			qid[2] = height;

			if (qid[0] > 100 || qid[0] < 0 || qid[1] > 150 || qid[1] < 0
					|| qid[2] > 200 || qid[2] < 0)
				return null;

			Individual p = new Individual(pid, qid);

			/**
			 * put the one into voter list
			 */
			voters.put(new Integer(pid), p);

			int put = Math.abs(rd.nextInt()) % 10;

			if (put == 0)
				peopleList.add(p);

			rlist.add(p);

			return p;
		} catch (Exception e) {
			// e.printStackTrace();
			return null;
		}
		// }
	}

	/**
	 * Load generated RL data by me
	 */
	public void readRLData(String dataset) {

		/**
		 * set up all diseases
		 */
		try {

			String tuple = "";
			BufferedReader reader = new BufferedReader(new InputStreamReader(
					new FileInputStream(dataset)));

			int update_num = 0;
			/**
			 * read in all the tuples
			 */
			while ((tuple = reader.readLine()) != null) {

				String[] fields = tuple.split(" ");

				int pid = 0;
				try {
					pid = Integer.parseInt(fields[0]);
				} catch (Exception e) {
					// e.printStackTrace();
					// System.out.println("errors 2!");
					continue;
				}
				float QID[] = new float[fields.length - 1];
				// for (int i = 0; i < Constant.DIM; i++) {
				// System.out.println(fields[i]);
				// }
				try {
					for (int i = 1; i < Constant.DIM; i++) {
						/**
						 * parse the qid from data
						 */
						QID[i] = Float.parseFloat(fields[i]);
					}
					/**
					 * get the disease
					 */
					Individual fresh = new Individual(pid, QID);

					Individual p = (Individual) voters.get(new Integer(pid));

					if (!fresh.equal(p)) {
						/**
						 * update the QID
						 */
						p.setQID(fresh.getQID());
						// System.out.println("QID updated");
						update_num++;
					}
				} catch (Exception e) {
					e.printStackTrace();

				}
			}
			// System.out.println("total update on QID " + update_num);
			reader.close();

		} catch (Exception e) {
			e.printStackTrace();
			// return null;
		}
	}

	/**
	 * The list of diseases
	 */
	public void readMicroData(String dataset) {
		/**
		 * clear the list of individuals
		 */
		for (int i = 0; i < rlist.size(); i++) {
			Individual p = (Individual) rlist.get(i);
			p.clearDisease();
			/**
			 * set the person to be not available
			 */
			p.setAvailble(true);
		}

		for (int i = 0; i < peopleList.size(); i++) {
			Individual p = (Individual) peopleList.get(i);
			p.clearDisease();
			/**
			 * set the person to be not available
			 */
			p.setAvailble(true);
		}
		/**
		 * set up all diseases
		 */
		try {
			String tuple = "";
			int size = 0;
			BufferedReader reader = new BufferedReader(new InputStreamReader(
					new FileInputStream(dataset)));
			while ((tuple = reader.readLine()) != null)
				size++;

			// System.out.println(size);
			reader.close();
			System.gc();

			reader = new BufferedReader(new InputStreamReader(
					new FileInputStream(dataset)));

			int num_of_tuples = 0;
			/**
			 * read in all the tuples
			 */
			while ((tuple = reader.readLine()) != null) {

				String[] fields = tuple.split("!");

				if (fields.length < 3) {
					System.out.println("errors!");
					System.out.println(fields[0]);
					continue;
				}

				int pid = 0;
				try {
					pid = Integer.parseInt(fields[1]);
				} catch (Exception e) {
					e.printStackTrace();
					System.out.println("errors 2!");
					continue;
				}
				/**
				 * get the disease
				 */
				String dis = fields[2].trim();

				Individual p = (Individual) voters.get(new Integer(pid));
				if (p == null) {
					System.out.println("null person: no entry in RL! ");
					continue;
				}
				/**
				 * if p is firstly added
				 */
				if (!peopleList.contains(p)) {
					peopleList.add(p);
					p.clearDisease();
				}
				/**
				 * set the holder of permanent sensitive values find the id of
				 * CANCER
				 */
				if (dis.indexOf("BREAST CANCER") >= 0) {
					p.addDisease(new Integer(Constant.BREAST_CANCER));
					// p.setRole(this, Constant.BREAST_CANCER, Constant.HOLDER,
					// Constant.CLEAN);

					// this.putIntoCohort(p, Constant.BREAST_CANCER, 0);
				}
				/**
				 * set the holder of civity cancer
				 */
				else if (dis.indexOf("CAVITY CANCER") >= 0) {
					p.addDisease(new Integer(Constant.CAVITY_CANCER));
					// p.setRole(this, Constant.CAVITY_CANCER, Constant.HOLDER,
					// Constant.CLEAN);
					// this.putIntoCohort(p, Constant.CAVITY_CANCER, 0);
				}
				/**
				 * find a diabete DIABETE
				 */
				else if (dis.indexOf("DIABETE") >= 0) {
					p.addDisease(new Integer(Constant.DIABETE));
					// p.setRole(this, Constant.DIABETE, Constant.HOLDER,
					// Constant.CLEAN);
					// this.putIntoCohort(p, Constant.DIABETE, 0);
				}
				/**
				 * set other disease
				 */
				else {
					Integer diseaseID = (Integer) SensitiveValue.diseaseToID
							.get(dis);
					p.addDisease(diseaseID);
				}
				num_of_tuples++;
			}
			System.out.println("number of tuples: " + num_of_tuples);
			reader.close();
			int count = 0;
			for (int i = 0; i < peopleList.size(); i++) {
				Individual person = (Individual) peopleList.get(i);
				count += person.getDisease().size();
			}
			System.out.println("final count: " + count);
		} catch (Exception e) {
			e.printStackTrace();
			// return null;
		}

	}

	/**
	 * preprocess the data
	 * 
	 * @param dataset
	 */
	public void preProcessData(String dataset) {

		/**
		 * clear the list of individuals
		 */
		for (int i = 0; i < rlist.size(); i++) {
			Individual p = (Individual) rlist.get(i);
			p.clearDisease();
		}

		/**
		 * set up all disease
		 */
		try {
			String tuple = "";

			// String[] results = new String[size];
			/**
			 * buffered reader
			 */
			BufferedReader reader = new BufferedReader(new InputStreamReader(
					new FileInputStream(dataset)));

			PrintWriter writer = new PrintWriter(new OutputStreamWriter(
					new FileOutputStream("re-action.txt")));
			/**
			 * read in all the tuples
			 */
			while ((tuple = reader.readLine()) != null) {

				String[] fields = tuple.split("!");
				// System.out.println(fields[0]);

				if (fields.length < 3) {
					// System.out.println("errors!");
					// System.out.println(fields[0]);
					continue;
				}

				int pid = 0;
				try {
					pid = Integer.parseInt(fields[1]);
				} catch (Exception e) {
					// e.printStackTrace();
					// System.out.println("errors 2!");
					continue;
				}
				/**
				 * get the disease
				 */
				String dis = fields[2];

				// System.out.println(fields[1] + " " + fields[2]);

				// find the individual and update his disease
				Individual p = (Individual) voters.get(new Integer(pid));
				if (p != null) {
					// p.addDisease(dis);
					writer.println(tuple);
					writer.flush();
				}
			}
			writer.close();
			// System.out.println(dis_no);
			reader.close();

		} catch (Exception e) {
			e.printStackTrace();
			// return null;
		}
	}

	/**
	 * assign a container for an individual
	 * 
	 * @return
	 */
	public SContainer assignContainer(Individual in, int dis) {

		/**
		 * if individual in is already in some container
		 */
		if (in.getContainer(dis) != null)
			return in.getContainer(dis);

		int position = rlist.indexOf(in);

		/**
		 * the search space
		 */
		int lower = position - Constant.CONTAINER_SIZE * 40;

		int upper = position + Constant.CONTAINER_SIZE * 40;

		if (lower < 0)
			lower = 0;

		if (upper > rlist.size())
			upper = rlist.size();

		List people = new Vector();

		List candidate = new Vector();
		/**
		 * add one into the people's list
		 */
		for (int i = lower; i < upper; i++)
			if (i != position)
				people.add(rlist.get(i));

		/**
		 * put clean individuals into the container
		 */
		for (int i = 0; i < Constant.CONTAINER_SIZE - 1; i++) {
			int role = Constant.CLEAN;
			Individual person = null;
			do {
				/**
				 * search NN while the person's role != Clean
				 */
				person = (Individual) this.findNNContainer(in, people);
				role = person.getRole(dis);
				if (role != Constant.CLEAN) {
					people.remove(person);
				}
				// System.out.println("find NN");
			} while (role != Constant.CLEAN);
			/**
			 * add the person into candidate list, and set its role as buddy
			 */
			candidate.add(person);
			/**
			 * set those new clean individuals' role to buddy
			 */
			person.setRole(this, dis, Constant.BUDDY, Constant.CLEAN);
			/**
			 * remove the selected buddy from people list
			 */
			people.remove(person);
		}

		/**
		 * find an absent buddy and add it into the container if necessary
		 */
		boolean added = false;
		do {
			Individual absentBuddy = findNN(in, absent);

			int role = absentBuddy.getRole(dis);
			System.out.println("in loop absent");
			/**
			 * if the role is not clean
			 */
			if (role != Constant.CLEAN && !candidate.contains(absentBuddy)) {
				absent.remove(absentBuddy);
				System.out.println("in loop present");
				continue;
			}

			if (role != Constant.CLEAN && candidate.contains(absentBuddy)) {
				absent.remove(absentBuddy);
				break;
			}

			/**
			 * if the candidate list not contain the absent clean
			 */
			if (!candidate.contains(absentBuddy)) {
				candidate.add(absentBuddy);
				/**
				 * set the role to buddy
				 */
				absentBuddy.setRole(this, dis, Constant.BUDDY, Constant.CLEAN);
				absent.remove(absentBuddy);
				added = true;
			}
		} while (added == false);

		/**
		 * find an absent buddy and add it into the container if necessary
		 */
		added = false;
		do {
			Individual presentBuddy = findNN(in, present);
			int role = presentBuddy.getRole(dis);

			/**
			 * if the role is not clean
			 */
			if (role != Constant.CLEAN && !candidate.contains(presentBuddy)) {
				present.remove(presentBuddy);
				System.out.println("in loop present");
				continue;
			}

			if (role != Constant.CLEAN && candidate.contains(presentBuddy)) {
				present.remove(presentBuddy);
				break;
			}
			/**
			 * if the candidate list not contain the present clean
			 */
			if (!candidate.contains(presentBuddy)) {
				candidate.add(presentBuddy);
				/**
				 * set the role to buddy
				 */
				presentBuddy.setRole(this, dis, Constant.BUDDY, Constant.CLEAN);
				present.remove(presentBuddy);
				added = true;
			}
		} while (added == false);

		/**
		 * create a new scontainer
		 */
		SContainer cn = new SContainer(dis, candidate);
		/**
		 * set the individual in a container
		 */
		in.setContainer(cn, dis);

		for (int m = 0; m < candidate.size(); m++) {
			Individual p = (Individual) candidate.get(m);
			p.setContainer(cn, dis);
		}
		return cn;
	}

	/**
	 * reset the perturbation count to 0
	 */
	public static void resetPerturbation() {
		replace = 0;
	}

	/**
	 * update all signatures in each anonymized groups
	 */
	public void updateAllSignatures() {
		for (int i = 0; i < groups.size(); i++) {
			AGroup ag = (AGroup) groups.get(i);
			updateSignatures(ag);
		}
	}

	/**
	 * @return the rate of perturbation
	 */
	public double getPerturbationRate() {
		return (double) replace / (double) rlist.size();
	}

	/**
	 * divide the RL into present and absent
	 */
	public void divideRL() {
		/**
		 * clear the list of present and absent
		 */
		absent.clear();
		present.clear();
		/**
		 * divide people into the 2 lists
		 */
		for (int i = 0; i < peopleList.size(); i++) {
			Individual person = (Individual) peopleList.get(i);
			if (person.getDisease().size() == 0) {
				boolean clean = true;
				for (int j = 0; j < Constant.num_of_ps; j++) {
					if (person.getRole(j) != Constant.CLEAN)
						clean = false;
				}
				/**
				 * if no disease, put the person into absent list
				 */
				if (clean == true)
					absent.add(person);
			} else {
				boolean clean = true;
				for (int j = 0; j < Constant.num_of_ps; j++) {
					if (person.getRole(j) != Constant.CLEAN)
						clean = false;
				}
				/**
				 * if the person has some disease, put the person into present
				 * list
				 */
				if (clean == true)
					present.add(person);
			}
		}

		System.out.println("!!!number of present individuals: "
				+ present.size());
		System.out.println("!!!number of absent individuals: " + absent.size());
		log.println("!!!number of present individuals: " + present.size());
		log.println("!!!number of present individuals: " + present.size());
	}
}

class Count {
	public int count = 1;

	public void increaseOne() {
		count++;
	}
}
