package repub;

//#define	DIM 5

/*
 * An Hcode holds the Hilbert Curve sequence number of a point as an array
 * of unsigned ints. The least significant bit of hcode[0] is the least
 * significant bit of the sequence number.
 */

class Hcode {
	static int DIM = 5;

	public int hcode[] = new int[DIM];

	Hcode() {
		for (int i = 0; i < hcode.length; i++) {
			hcode[i] = 0;
		}
	}

	Hcode(Hcode h) {
		for (int i = 0; i < hcode.length; i++) {
			hcode[i] = h.hcode[i];
		}
	}
};

public class Hilbert {
	// typedef Hcode Point;

	/*
	 * g_mask can be omitted... g_mask[x] can be replaced by (1 << DIM - 1 - x)
	 */
	static int g_mask[] = { 16, 8, 4, 2, 1 };

	/*
	 * retained for historical reasons: the number of bits in an attribute
	 * value: effectively the order of a curve
	 */
	static int NUMBITS = 31;

	/*
	 * the number of bits in a word used to store an hcode (or in an element of
	 * an array that's used)
	 */
	static int WORDBITS = 31;

	static Hcode negate(Hcode para, int index) {
		Hcode Z = new Hcode(para);

		Hcode returnValue = new Hcode();

		int i;

		int arrayNo;
		int bitNo;

		int temp;

		int mask;

		for (i = 0; i < returnValue.DIM; i++) {
			returnValue.hcode[i] = 0;
		}

		arrayNo = index / NUMBITS;
		bitNo = index % NUMBITS;

		for (i = 0; i < arrayNo; i++) {
			returnValue.hcode[i] = ~(Z.hcode[i]);
		}

		// create mask for Most Significant byte
		mask = 0;
		for (i = 0; i <= bitNo; i++) {
			mask |= (1 << i);
		}
		//
		temp = (~(Z.hcode[arrayNo]) & mask);
		returnValue.hcode[arrayNo] = temp;

		return returnValue;
	}

	// add a bit at the index-th position
	static Hcode add_bit(Hcode para, int index) {
		Hcode Z = new Hcode(para);

		Hcode returnValue = new Hcode();

		int i;

		int arrayNo;
		int bitNo;

		int temp;

		int mask;

		for (i = 0; i < returnValue.DIM; i++) {
			returnValue.hcode[i] = 0;
		}

		arrayNo = index / NUMBITS;
		bitNo = index % NUMBITS;

		for (i = 0; i <= arrayNo; i++) {
			returnValue.hcode[i] = Z.hcode[i];
		}

		// add one at the index-th position (i.e. bitNo-th position)
		returnValue.hcode[arrayNo] += (1 << bitNo);

		// create mask for Most Significant byte
		mask = 0;
		for (i = 0; i <= bitNo; i++) {
			mask |= (1 << i);
		}
		//
		temp = (returnValue.hcode[arrayNo] & mask);
		returnValue.hcode[arrayNo] = temp;

		return returnValue;
	}

	// ------------------------------------------------

	/* ============================================================================ */
	/* H_encode */
	/* ============================================================================ */
	/*
	 * given the coordinates of a point, it finds the sequence number of the
	 * point on the Hilbert Curve
	 */
	static Hcode H_encode(Hcode para) {
		Hcode p = new Hcode(para);
		int mask = (int) 1 << WORDBITS - 1, element, temp1, temp2, A, W = 0, S, tS, T, tT, J, P = 0, xJ;
		Hcode h = new Hcode();
		int i = NUMBITS * Hcode.DIM - Hcode.DIM, j;

		for (j = A = 0; j < Hcode.DIM; j++)
			if ((p.hcode[j] & mask) != 0)
				A |= g_mask[j];

		S = tS = A;

		P |= S & g_mask[0];
		for (j = 1; j < Hcode.DIM; j++)
			if ((S & g_mask[j] ^ (P >> 1) & g_mask[j]) != 0)
				P |= g_mask[j];

		/* add in DIM bits to hcode */
		element = i / WORDBITS;
		if (i % WORDBITS > WORDBITS - Hcode.DIM) {
			h.hcode[element] |= P << i % WORDBITS;
			// *** Ray changed
			// h.hcode[element + 1] |= P >> WORDBITS - i % WORDBITS;
			h.hcode[element + 1] |= P >> (WORDBITS - i % WORDBITS);
		} else
			// *** Ray changed
			// h.hcode[element] |= P << i - element * WORDBITS;
			h.hcode[element] |= P << (i - element * WORDBITS);

		J = Hcode.DIM;
		for (j = 1; j < Hcode.DIM; j++)
			if ((P >> j & 1) == (P & 1))
				continue;
			else
				break;
		if (j != Hcode.DIM)
			J -= j;
		xJ = J - 1;

		if (P < 3)
			T = 0;
		else if ((P % 2) != 0)
			T = (P - 1) ^ (P - 1) / 2;
		else
			T = (P - 2) ^ (P - 2) / 2;
		tT = T;

		for (i -= Hcode.DIM, mask >>= 1; i >= 0; i -= Hcode.DIM, mask >>= 1) {
			for (j = A = 0; j < Hcode.DIM; j++)
				if ((p.hcode[j] & mask) != 0)
					A |= g_mask[j];

			W ^= tT;
			tS = A ^ W;
			if (xJ % Hcode.DIM != 0) {
				temp1 = tS << xJ % Hcode.DIM;
				// *** Ray changed
				// temp2 = tS >> DIM - xJ % DIM;
				temp2 = tS >> (Hcode.DIM - xJ % Hcode.DIM);
				S = temp1 | temp2;
				S &= ((int) 1 << Hcode.DIM) - 1;
			} else
				S = tS;

			P = S & g_mask[0];
			for (j = 1; j < Hcode.DIM; j++)
				if ((S & g_mask[j] ^ (P >> 1) & g_mask[j]) != 0)
					P |= g_mask[j];

			/* add in DIM bits to hcode */
			element = i / WORDBITS;
			if (i % WORDBITS > WORDBITS - Hcode.DIM) {
				h.hcode[element] |= P << i % WORDBITS;
				// *** Ray changed
				// h.hcode[element + 1] |= P >> WORDBITS - i % WORDBITS;
				h.hcode[element + 1] |= P >> (WORDBITS - i % WORDBITS);
			} else
				// *** Ray changed
				// h.hcode[element] |= P << i - element * WORDBITS;
				h.hcode[element] |= P << (i - element * WORDBITS);

			if (i > 0) {
				if (P < 3)
					T = 0;
				else if ((P % 2) != 0)
					T = (P - 1) ^ (P - 1) / 2;
				else
					T = (P - 2) ^ (P - 2) / 2;

				if (xJ % Hcode.DIM != 0) {
					temp1 = T >> xJ % Hcode.DIM;
					// *** Ray changed
					// temp2 = T << DIM - xJ % DIM;
					temp2 = T << (Hcode.DIM - xJ % Hcode.DIM);
					tT = temp1 | temp2;
					tT &= ((int) 1 << Hcode.DIM) - 1;
				} else
					tT = T;

				J = Hcode.DIM;
				for (j = 1; j < Hcode.DIM; j++)
					if ((P >> j & 1) == (P & 1))
						continue;
					else
						break;
				if (j != Hcode.DIM)
					J -= j;

				xJ += J - 1;
				/* J %= DIM; */
			}
		}
		return h;
	}

	/* ============================================================================ */
	/* H_decode */
	/* ============================================================================ */
	/*
	 * given the sequence number of a point, it finds the coordinates of the
	 * point on the Hilbert Curve
	 */
	static Hcode H_decode(Hcode H) {
		int mask = (int) 1 << WORDBITS - 1, element, temp1, temp2, A, W = 0, S, tS, T, tT, J, P = 0, xJ;
		Hcode pt = new Hcode();
		int i = NUMBITS * Hcode.DIM - Hcode.DIM, j;

		/*--- P ---*/
		element = i / WORDBITS;
		P = H.hcode[element];
		if (i % WORDBITS > WORDBITS - Hcode.DIM) {
			temp1 = H.hcode[element + 1];
			P >>= i % WORDBITS;
			temp1 <<= WORDBITS - i % WORDBITS;
			P |= temp1;
		} else
			P >>= i % WORDBITS; /* P is a DIM bit hcode */

		/* the & masks out spurious highbit values */
		// #if DIM < WORDBITS
		// P &= (1 << DIM) -1;
		// #endif
		/*--- xJ ---*/
		J = Hcode.DIM;
		for (j = 1; j < Hcode.DIM; j++)
			if ((P >> j & 1) == (P & 1))
				continue;
			else
				break;
		if (j != Hcode.DIM)
			J -= j;
		xJ = J - 1;

		/*--- S, tS, A ---*/
		A = S = tS = P ^ P / 2;

		/*--- T ---*/
		if (P < 3)
			T = 0;
		else if ((P % 2) != 0)
			T = (P - 1) ^ (P - 1) / 2;
		else
			T = (P - 2) ^ (P - 2) / 2;

		/*--- tT ---*/
		tT = T;

		/*--- distrib bits to coords ---*/
		for (j = Hcode.DIM - 1; P > 0; P >>= 1, j--)
			if ((P & 1) != 0)
				pt.hcode[j] |= mask;

		for (i -= Hcode.DIM, mask >>= 1; i >= 0; i -= Hcode.DIM, mask >>= 1) {
			/*--- P ---*/
			element = i / WORDBITS;
			P = H.hcode[element];
			if (i % WORDBITS > WORDBITS - Hcode.DIM) {
				temp1 = H.hcode[element + 1];
				P >>= i % WORDBITS;
				temp1 <<= WORDBITS - i % WORDBITS;
				P |= temp1;
			} else
				P >>= i % WORDBITS; /* P is a DIM bit hcode */

			/* the & masks out spurious highbit values */
			// #if DIM < WORDBITS
			// P &= (1 << DIM) -1;
			// #endif
			/*--- S ---*/
			S = P ^ P / 2;

			/*--- tS ---*/
			if (xJ % Hcode.DIM != 0) {
				temp1 = S >> xJ % Hcode.DIM;
				// *** Ray changed
				// temp2 = S << DIM - xJ % DIM;
				temp2 = S << (Hcode.DIM - xJ % Hcode.DIM);
				tS = temp1 | temp2;
				tS &= ((int) 1 << Hcode.DIM) - 1;
			} else
				tS = S;

			/*--- W ---*/
			W ^= tT;

			/*--- A ---*/
			A = W ^ tS;

			/*--- distrib bits to coords ---*/
			for (j = Hcode.DIM - 1; A > 0; A >>= 1, j--)
				if ((A & 1) != 0)
					pt.hcode[j] |= mask;

			if (i > 0) {
				/*--- T ---*/
				if (P < 3)
					T = 0;
				else if ((P % 2) != 0)
					T = (P - 1) ^ (P - 1) / 2;
				else
					T = (P - 2) ^ (P - 2) / 2;

				/*--- tT ---*/
				if (xJ % Hcode.DIM != 0) {
					temp1 = T >> xJ % Hcode.DIM;
					// *** Ray changed
					// temp2 = T << DIM - xJ % DIM;
					temp2 = T << (Hcode.DIM - xJ % Hcode.DIM);
					tT = temp1 | temp2;
					tT &= ((int) 1 << Hcode.DIM) - 1;
				} else
					tT = T;

				/*--- xJ ---*/
				J = Hcode.DIM;
				for (j = 1; j < Hcode.DIM; j++)
					if ((P >> j & 1) == (P & 1))
						continue;
					else
						break;
				if (j != Hcode.DIM)
					J -= j;
				xJ += J - 1;
			}
		}
		return pt;
	}

	static void printPoint(Hcode para) {
		Hcode pt = new Hcode(para);
		int i;

		System.out.print("(");
		for (i = 0; i < Hcode.DIM; i++) {
			System.out.print(pt.hcode[i]);

			if (i != Hcode.DIM - 1) {
				System.out.print(", ");
			}
		}
		System.out.print(")\n");

	}

	static int testShift() {
		int no;
		int no2;

		no = 4;

		no2 = no << 2;

		System.out.println("no  = " + no);
		System.out.println("no2 = " + no2);

		return 0;
	}

	static int doHilbert(Individual p) {
		Hcode multiDim = new Hcode();
		float[] qid = p.getQID();
		Hcode oneDim;

		float[] q = new float[qid.length];
		for (int i = 0; i < qid.length; i++) {
			q[i] = qid[i];
		}

		for (int i = 0; i < Constant.DIM; i++)
			multiDim.hcode[i] = (int) q[i];

		oneDim = H_encode(multiDim);
		// p.setHashValue(oneDim.hcode[0]);
		return oneDim.hcode[0];
	}

	static int doHilbert(Hcode para) {
		Hcode multiDim = new Hcode(para);
		Hcode oneDim;

		oneDim = H_encode(multiDim);
		return oneDim.hcode[0];
	}

	public static void main(String[] args) {
		Hcode multiDim = new Hcode();
		// Hcode oneDim;

		multiDim.hcode[0] = 2;
		multiDim.hcode[1] = 4;
		multiDim.hcode[2] = 1;
		multiDim.hcode[3] = 2;
		multiDim.hcode[4] = 1;

		Hilbert hil = new Hilbert();
		System.out.println(hil.doHilbert(multiDim));
	}
}