package repub;

import java.io.Serializable;
import java.util.List;
import java.util.Vector;

public class Individual implements Serializable, Comparable {

	/**
	 * indicate whether the individual is extinct
	 */
	private boolean extinct = false;

	/**
	 * the individual's ID
	 */
	private int PID = 0;

	/**
	 * the associated containers
	 */
	private SContainer[] parent = new SContainer[Constant.num_of_ps];

	/**
	 * the QID field
	 */
	private float[] QID = null;

	/**
	 * the hash value
	 */
	private int hashValue = 0;

	/**
	 * the disease that the individual has
	 */
	private List diseases = new Vector();

	/**
	 * its role for every permanent sensitive values, also the signatures
	 */
	private int[] roles = new int[Constant.num_of_ps];

	/**
	 * if it is avaible to form a clique
	 */
	private boolean available = true;

	private boolean inContainer = false;

	private ICohort[] owners = new ICohort[Constant.L];

	/**
	 * initialize indivudal
	 */
	public Individual() {
		/**
		 * set containers
		 */
		for (int i = 0; i < parent.length; i++)
			parent[i] = null;
	}

	/**
	 * check if the individuals have been in container
	 * 
	 * @return
	 */
	public boolean getInContainer() {
		return inContainer;
	}

	/**
	 * set whether the individual is in a container
	 * 
	 * @param in
	 */
	public void setInContainer(boolean in) {
		inContainer = in;
	}

	/**
	 * constructor
	 * 
	 * @param q
	 */
	public Individual(int pid, float[] q) {
		PID = pid;
		QID = q;
		for (int i = 0; i < roles.length; i++)
			roles[i] = Constant.CLEAN;
		/**
		 * get the hash value by hilbert curve
		 */
		Hcode multiDim = new Hcode();
		for (int i = 0; i < Constant.DIM; i++)
			multiDim.hcode[i] = (int) QID[i];
		hashValue = Hilbert.doHilbert(multiDim);

		/**
		 * set containers
		 */
		for (int i = 0; i < parent.length; i++)
			parent[i] = null;
	}

	/**
	 * get the QID fields
	 * 
	 * @return
	 */
	public float[] getQID() {
		return QID;
	}

	/**
	 * randomly change the QIDs in the domain
	 */
	public void randomChangeQID() {

	}

	/**
	 * get the container that the individual owns
	 * 
	 * @return
	 */
	public SContainer getContainer(int dis) {
		return parent[dis];
	}

	/**
	 * set the container of disease dis
	 * 
	 * @param cn
	 */
	public void setContainer(SContainer cn, int dis) {
		parent[dis] = cn;
	}

	/**
	 * set the value to be available or not
	 * 
	 * @param av
	 */
	public void setAvailble(boolean av) {
		available = av;
	}

	/**
	 * return whether the individual is available to be included in a AG
	 * 
	 * @return
	 */
	public boolean isAvailable() {
		return available;
	}

	/**
	 * print the individual
	 */
	public String toString() {
		String s = Integer.toString(this.PID);
		for (int i = 0; i < Constant.DIM; i++) {
			s = s + " " + QID[i];
		}
		// for (int i = 0; i < diseases.size(); i++)
		// s = s + " " + diseases.get(i);
		// for (int i = 0; i < roles.length; i++)
		// s = s + " " + roles[i];
		return s;
	}

	public int getHashValue() {
		return hashValue;
	}

	public void addDisease(Integer it) {
		// if (!diseases.(s))
		// for (int i = 0; i < diseases.size(); i++) {
		// String dis = (String) diseases.get(i);
		// if (dis.equals(s))
		// return;
		// }
		if (!diseases.contains(it))
			diseases.add(it);
		else {
			// System.out.println("repeated disease: " + this.PID);
		}
	}

	public void clearDisease() {
		diseases.clear();
	}

	public void setQID(float[] qid) {
		this.QID = qid;
	}

	/**
	 * get the person's ID
	 * 
	 * @return
	 */
	public int getPID() {
		return PID;
	}

	/**
	 * check whether two objects are identical objects
	 * 
	 * @param o
	 * @return
	 */
	public boolean equal(Object o) {
		Individual in = (Individual) o;
		if (this.getPID() != in.getPID())
			return false;
		// for (int i = 0; i < QID.length; i++) {
		// if (QID[i] != in.QID[i])
		// return false;
		// }
		return true;
	}

	/**
	 * get the Lp^2 distance with an other individual
	 * 
	 * @param other
	 * @return
	 */
	public float distance(Individual other) {
		float distance = 0;
		float[] qid2 = other.getQID();
		for (int i = 0; i < QID.length; i++) {
			distance += (this.QID[i] - qid2[i]) * (this.QID[i] - qid2[i]);
		}
		return distance;
	}

	public void setHashValue(int code) {
		hashValue = code;
	}

	/**
	 * get the role of disease id
	 * 
	 * @param diseaseid
	 * @return
	 */
	public int getRole(int diseaseid) {
		/**
		 * return the roles on attribute i
		 */
		return roles[diseaseid];
	}

	/**
	 * set the individual's role
	 */
	public void setRole(Generalization gn, int entry, int role, int fromrole) {
		this.roles[entry] = role;
		/**
		 * assign a container for that entry
		 */
		if (role == Constant.HOLDER || role == Constant.DECOY)
			parent[entry] = gn.assignContainer(this, entry);
	}

	public void setRole(Generalization gn, int dis, int role) {
		this.roles[dis] = role;
		/**
		 * assign a container for that entry
		 */
		if (role == Constant.HOLDER || role == Constant.DECOY)
			parent[dis] = gn.assignContainer(this, dis);
	}

	/**
	 * return the individual's disease list
	 * 
	 * @return
	 */
	public List getDisease() {
		return diseases;
	}

	/**
	 * compare based on hash value
	 * 
	 * @param o
	 * @return
	 */
	public int compareTo(Object o) {
		Individual in = (Individual) o;
		if (this.hashValue > in.hashValue)
			return 1;
		if (this.hashValue < in.hashValue)
			return -1;
		else
			return 0;
	}

	// public float[] getContainerQIDUpper() {
	// float[] value =
	// }

	/**
	 * get the lower bound of containers
	 */
	public float[] getContainerQIDLower() {
		float value[] = new float[Constant.DIM];
		for (int i = 0; i < value.length; i++)
			value[i] = this.QID[i];
		for (int s = 0; s < Constant.num_of_ps; s++) {
			if (parent[s] != null) {
				/**
				 * individual list
				 */
				List individuals = parent[s].getIndividuals();
				float lower[][] = new float[individuals.size()][];
				for (int i = 0; i < individuals.size(); i++) {
					Individual in = (Individual) individuals.get(i);
					lower[i] = in.getQID();
				}
				for (int j = 0; j < lower.length; j++) {
					for (int k = 0; k < Constant.DIM; k++) {
						if (value[k] > lower[j][k])
							value[k] = lower[j][k];
					}
				}
			}
		}
		return value;
	}

	/**
	 * get the upper bound of containers
	 * 
	 * @return
	 */
	public float[] getContainerQIDUpper() {
		float value[] = new float[Constant.DIM];
		for (int i = 0; i < value.length; i++)
			value[i] = this.QID[i];
		for (int s = 0; s < Constant.num_of_ps; s++) {
			if (parent[s] != null) {
				/**
				 * individual list
				 */
				List individuals = parent[s].getIndividuals();
				float upper[][] = new float[individuals.size()][];
				for (int i = 0; i < individuals.size(); i++) {
					Individual in = (Individual) individuals.get(i);
					upper[i] = in.getQID();
				}
				for (int j = 0; j < upper.length; j++) {
					for (int k = 0; k < Constant.DIM; k++) {
						if (value[k] < upper[j][k])
							value[k] = upper[j][k];
					}
				}
			}
		}
		return value;
	}

	/**
	 * check if the individual has contracted disease dis
	 * 
	 * @param dis
	 * @return
	 */
	public boolean isContract(Integer dis) {
		for (int i = 0; i < diseases.size(); i++) {
			Integer dis2 = (Integer) diseases.get(i);
			if (dis.equals(dis2))
				return true;
		}
		return false;
	}

	/**
	 * remove a disease dis from the individual's disease list
	 * 
	 * @param dis
	 */
	public void removeDisease(Integer dis) {
		for (int i = 0; i < diseases.size(); i++) {
			Integer dis2 = (Integer) diseases.get(i);
			if (dis.equals(dis2))
				diseases.remove(i);
		}
	}

	/**
	 * set the cohort
	 * 
	 * @param cohort
	 */
	public void setCohort(ICohort cohort, int dis) {
		owners[dis] = cohort;
	}

	/**
	 * get the cohort
	 * 
	 * @return
	 */
	public ICohort getCohort(int dis) {
	 return owners[dis];
	}
	/**
	 * check whether the individual is extinct
	 * 
	 * @return
	 */
	public boolean isExtinct() {
		return extinct;
	}

	/**
	 * set whether the individual is extinct
	 * 
	 * @param ext
	 */
	public void setExtinct(boolean ext) {
		extinct = ext;
	}
}