package repub;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

public class ResultGenerator {

	public static void main(String[] args) {
		try {
			TestError.initialTestEnvironment();
			/**
			 * report the time used for each generalization process
			 */
			PrintWriter time = new PrintWriter(new OutputStreamWriter(
					new FileOutputStream("timeL.txt")));
			/**
			 * report the perturbation rate for each generalization process
			 */
			PrintWriter perturbation = new PrintWriter(new OutputStreamWriter(
					new FileOutputStream("replaceL.txt")));

			/**
			 * report the perturbation rate for each generalization process
			 */
			PrintWriter error = new PrintWriter(new OutputStreamWriter(
					new FileOutputStream("errorL.txt")));

			/**
			 * reset perturbation rate to zero
			 */
			Generalization.resetPerturbation();

			/**
			 * change L and generate
			 */
			for (int l = 3; l <= 8; l = l + 2) {
				/**
				 * set the parameter L for L-scarcity
				 */
				Constant.L = l;
				Generalization gn = new Generalization(
						"publication/composition" + 2 + ".txt", 2);
				gn.readStartRLData("data/REPORTS.TXT");

				/**
				 * read microdata to initiate sensitive values
				 */
				SensitiveValue.readMicroData("data/re-action.txt");
				
				time.print(l);
				perturbation.print(l);
				error.print(l);

				/**
				 * publish 19 rounds
				 */
				for (int i = 2; i < 21; i++) {
					long start = System.currentTimeMillis();
					File folder = new File(".//publicationL" + Constant.L);

					folder.mkdir();

					gn = new Generalization("publicationL" + l + "/composition"
							+ i + ".txt", i);
					gn.loadRLData();

					gn.readMicroData("microdata/T" + i + ".txt");
					gn.topDownPartition();

					/**
					 * update all signatures of individuals
					 */
					// gn.updateAllSignatures();
					long end = System.currentTimeMillis();

					if (i % 4 == 0) {
						time.print(" " + (end - start));
						perturbation.print(" " + gn.getPerturbationRate());
						error.print(" "
								+ TestError.unitTestError("microdata/T" + i
										+ ".txt", "publicationL" + l
										+ "/composition" + i + ".txt"));

						time.flush();
						perturbation.flush();
						error.flush();
					}
				}
				time.println();
				perturbation.println();
				error.flush();
			}
			time.close();
			perturbation.close();
			error.close();

			/**
			 * restore L to 3, and test for different parameter c
			 */
			Constant.L = 3;
			/**
			 * the two files for parameter C, the size of containers
			 */
			time = new PrintWriter(new OutputStreamWriter(new FileOutputStream(
					"timeC.txt")));
			perturbation = new PrintWriter(new OutputStreamWriter(
					new FileOutputStream("replaceC.txt")));
			error = new PrintWriter(new OutputStreamWriter(
					new FileOutputStream("errorC.txt")));

			/**
			 * reset accumulate perturbation rate to zero
			 */
			Generalization.resetPerturbation();
			/**
			 * test for different parameter C
			 */
			for (int c = 3; c <= 8; c += 2) {
				Constant.CONTAINER_SIZE = c;
				File folder = new File(".//publicationC"
						+ Constant.CONTAINER_SIZE);
				folder.mkdir();

				/**
				 * initialize a generalization
				 */
				Generalization gn = new Generalization(
						"publication/composition" + 2 + ".txt", 2);
				gn.readStartRLData("data/REPORTS.TXT");

				SensitiveValue.readMicroData("data/re-action.txt");

				time.print(c);
				perturbation.print(c);
				error.print(c);
				
				/**
				 * publish data for 19 rounds
				 */
				for (int i = 2; i < 21; i++) {
					long start = System.currentTimeMillis();
					gn = new Generalization("publicationC"
							+ Constant.CONTAINER_SIZE + "/composition" + i
							+ ".txt", i);

					gn.loadRLData();

					gn.readMicroData("microdata/T" + i + ".txt");
					gn.topDownPartition();

					/**
					 * update all signatures of individuals updated in
					 * topDownPartition()
					 */
					// gn.updateAllSignatures();
					long end = System.currentTimeMillis();

					if (i % 4 == 0) {
						time.println(" " + (end - start));
						perturbation.println(" " + gn.getPerturbationRate());
						error.print(" "
								+ TestError.unitTestError("microdata/T" + i
										+ ".txt", "publicationC" + c
										+ "/composition" + i + ".txt"));

						time.flush();
						perturbation.flush();
						error.flush();
					}
				}
				time.println();
				perturbation.println();
				error.println();
			}

			time.close();
			perturbation.close();
			error.close();

			Constant.L = 3;
			Constant.CONTAINER_SIZE = 3;

			/**
			 * the result of time and perturbation rate
			 */
			time = new PrintWriter(new OutputStreamWriter(new FileOutputStream(
					"timeK.txt")));
			perturbation = new PrintWriter(new OutputStreamWriter(
					new FileOutputStream("replaceK.txt")));
			error = new PrintWriter(new OutputStreamWriter(
					new FileOutputStream("errorK.txt")));

			Generalization.resetPerturbation();

			/**
			 * size of participation knowledge
			 */
			for (int K = 3; K <= 8; K += 2) {
				Constant.CONTAINER_SIZE = Constant.CONTAINER_SIZE + K;

				File folder = new File(".//publicationK"
						+ Constant.CONTAINER_SIZE);

				folder.mkdir();

				Generalization gn = new Generalization(
						"publication/composition" + 2 + ".txt", 2);
				gn.readStartRLData("data/REPORTS.TXT");

				SensitiveValue.readMicroData("data/re-action.txt");

				time.print(K);
				perturbation.print(K);
				error.print(K);

				for (int i = 2; i < 21; i++) {
					long start = System.currentTimeMillis();
					gn = new Generalization("publicationK"
							+ Constant.CONTAINER_SIZE + "/composition" + i
							+ ".txt", i);

					gn.loadRLData();
					gn.readMicroData("microdata/T" + i + ".txt");
					gn.topDownPartition();

					/**
					 * update all signatures of individuals updated in top down
					 * partition
					 */
					// gn.updateAllSignatures();
					long end = System.currentTimeMillis();

					/**
					 * print the running time and perturbation
					 */
					if (i % 4 == 0) {
						time.println(" " + (end - start));
						perturbation.println(" " + gn.getPerturbationRate());
						error.print(" "
								+ TestError.unitTestError("microdata/T" + i
										+ ".txt", "publicationK" + K
										+ "/composition" + i + ".txt"));
						time.flush();
						perturbation.flush();
						error.flush();
					}
				}
				time.println();
				perturbation.println();
				error.println();
			}

			time.close();
			perturbation.close();
			error.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
